;**********************************************************************
;
;    Interface de commande pour 8 servomoteurs,
;    contrl par une liaison srie
;
; La dure de l'impulsion servo = 2,473 - 0,008 * Position
; ( Pour Position compris entre 4 et 255 ) entre 0,433 et 2,441 ms.
; Si Position = 0 le servo est dsactiv ( pas d'impulsion gnr )
;
; La commande est reu par la liaison srie au format :
; 1 bit de start, 8 bit de donne, 1 bit de stop  2400 Bauds
;
; Pour dplacer un servo, 4 octets doit tre envoy :
;
;   1 - Entte = 255
;   2 - Adresse de la carte (valeur identique au port RA2  RA4 )
;   3 - Numro du servo ( de 0  7 )
;   4 - Position du servo ( de 4  255 ou 0 pour le dsactiver ) 
;
;     Pour le schma de la carte et d'autres informations,
;     voir la page :
;     http://fribotte.free.fr/bdtech/8servomoteurs/8servomoteurs.html
;
;**********************************************************************
;
;	Fichier:	8servos.asm
;	Date:		15/05/2001
;	Version:	V1.0
;
;	Auteur:		Jean-Claude
;	Compagnie:      http://fribotte.free.fr/
;
;**********************************************************************
;
;  Brochage du PIC sur la carte :
;
;  Port | Broche | E/S | Nom             | Description
;       |        |     |                 |
;  RA0  |    17  |  E  |                 | Libre
;  RA1  |    18  |  E  | CEAPortSerie    | Entre Srie
;  RA2  |     1  |  E  | CEAAdresse0     | Entre Adresse carte 0
;  RA3  |     2  |  E  | CEAAdresse1     | Entre Adresse carte 1
;  RA4  |     3  |  E  | CEAAdresse2     | Entre Adresse carte 2
;       |        |     |                 |
;  RB0  |     6  |  S  | CSBPortSM0      | Sorties servo moteur
;  RB1  |     7  |  S  | CSBPortSM1      | 
;  RB2  |     8  |  S  | CSBPortSM2      | 
;  RB3  |     9  |  S  | CSBPortSM3      |
;  RB4  |    10  |  S  | CSBPortSM4      | 
;  RB5  |    11  |  S  | CSBPortSM5      | 
;  RB6  |    12  |  S  | CSBPortSM6      | 
;  RB7  |    13  |  S  | CSBPortSM7      | 
;
;**********************************************************************
;
;	Notes:
;		16F84	Quartz 4Mhz
;
;**********************************************************************
;
;	Historique:    
;		15/05/2001	Cration
;
;**********************************************************************

	list      p=16F84             ; Type du processeur
	#include <p16F84.inc>         ; Dfinition des variables

; Pas de code protg
; Pas de watchdog timer
; Dlai de dmarrage
; Horloge XT.

	__CONFIG   _CP_OFF & _WDT_OFF & _PWRTE_ON & _XT_OSC


;**********************************************************************
;
;	Dclaration des variables et constantes ( de 0x0C  0X4F )
;
;**********************************************************************

w_temp        	EQU     0x0C        	; Sauvegarde contexte W 
status_temp   	EQU     0x0D        	; Sauvegarde contexte status

CEAPortSerie	EQU 	1
CEAAdresse0	EQU 	2
CEAAdresse1	EQU 	3
CEAAdresse2	EQU 	4

CSBPortSM0      EQU 	0
CSBPortSM1      EQU 	1
CSBPortSM2      EQU 	2
CSBPortSM3      EQU 	3
CSBPortSM4      EQU 	4
CSBPortSM5      EQU 	5
CSBPortSM6      EQU 	6
CSBPortSM7      EQU 	7

Phase		EQU	0x0E		; Les diffrents cycles.
CPhase1		EQU	0		; Codage par bit
CPhase2		EQU	1
CPhase3		EQU	2
CPhaseMax	EQU	3

CTMR1		EQU	D'206'
CPrecale1	EQU	B'10000010'

CPwmMin		EQU	D'4'		; Min de position

IndPwm		EQU	0x0F		; Numero du servo en cours de traitement
IndTmp		EQU	0x10
ValTmp		EQU	0x11
PwmC		EQU	0x12		; position du servo en cours de traitement
PwmSM0		EQU	0x13		; Position des servos
PwmSM1		EQU	0x14
PwmSM2		EQU	0x15
PwmSM3		EQU	0x16
PwmSM4		EQU	0x17
PwmSM5		EQU	0x18
PwmSM6		EQU	0x19
PwmSM7		EQU	0x1A

Donnee		EQU	0x20		; Octet reu par FGet
NbrDonnee	EQU	0x21
PauseV		EQU	0x22

AdrCarte	EQU	0x23		; Adresse de la carte ( RA2  RA4 )
AdrServo	EQU	0x24		; Numro du servo lu sur srie
ValServo	EQU	0x25		; Position lu sur srie

;**************************************************************** DEBUT
		ORG     0x000           ; Vecteur reset
  		GOTO    main            ; Dbut du programme

;**********************************************************************
;
;    Interruption TimeR
;
;**********************************************************************

		ORG     0x004           ; Vecteur d'interruption

; Les interruptions TMR sont utilises pour gre les signaux destins
; aux servomoteurs.
; Les signaux sont grs les un aprs les autres toutes les 20ms
; (soit 2,5 ms rserv pour chaque servo )
; A la premire interruption (Phase1),
; le port est mis  1 ( si le servo est active )  pendant 0,412 ms
; A la 2eme (Phase2) le port est toujours a 1
; entre 0,026 et 2,034 ms (proportionnel a la position courant )
; A la 3eme (Phase3) le port est mis a 0,
; pour une dure de 0,054  2,062 ms
; (pour que la dure totale des 3 phases soit toujours  2,5 ms )
; Et ainsi de suite pour les autres servos.
; La dure de l'impulsion = 2,473 - 0,008 * Position
; ( Pour Position compris entre 4 et 255 )
; Impulsion comprise entre 0,433 et 2,441 ms.
; Si Position = 0 le servo est dsactiv ( pas d'impulsion gnr )
;
; Taux d'occupation du cpu compris entre 4,32% et 5,28%.
; 30 ou 54 cycles en phase1
; 28 cycles en phase2 
; 50 cycles en phase3

;***** Sauvegarde contexte
		MOVWF   w_temp		; Sauve Registre W
		MOVF	STATUS,W		; Sauve Status
		MOVWF	status_temp		; /

		BCF 	INTCON,	T0IF	; On efface l'interruption TimeR

; Aiguillage vers les diffrentes phases

		BTFSC	Phase,CPhase3	; Attention Phase 2 et 3,
		GOTO	LPhase3		; on une dure minimale critique.
		BTFSC	Phase,CPhase2
		GOTO	LPhase2
		;GOTO	LPhase1

LPhase1
; Init du TMR0 (Dure fixe 0,412 ms)
		MOVLW	CTMR1
		MOVWF	TMR0

; Phase suivante
		MOVLW	1<<CPhase2
		MOVWF	Phase

; Lecture du Pwm courant

		MOVLW	PwmSM0
		ADDWF	IndPwm,W
		MOVWF	FSR
		MOVF	INDF,W
		MOVWF	PwmC

; Dbut de  l'impulsions du servo s'il est actif

		BTFSC	STATUS,Z
		GOTO	BSuite1

		MOVF	IndPwm,W
		MOVWF	IndTmp

		BTFSC	STATUS,Z
		BSF	PORTB, CSBPortSM0

		DECF	IndTmp,F
		BTFSC	STATUS,Z
		BSF	PORTB, CSBPortSM1

		DECF	IndTmp,F
		BTFSC	STATUS,Z
		BSF	PORTB, CSBPortSM2

		DECF	IndTmp,F
		BTFSC	STATUS,Z
		BSF	PORTB, CSBPortSM3

		DECF	IndTmp,F
		BTFSC	STATUS,Z
		BSF	PORTB, CSBPortSM4

		DECF	IndTmp,F
		BTFSC	STATUS,Z
		BSF	PORTB, CSBPortSM5

		DECF	IndTmp,F
		BTFSC	STATUS,Z
		BSF	PORTB, CSBPortSM6

		DECF	IndTmp,F
		BTFSC	STATUS,Z
		BSF	PORTB, CSBPortSM7

BSuite1
		GOTO	LFinPhase

LPhase2
; Dure de 0,026 ms  2,034 ms selon le Pwm courant
		nop
		nop
		nop
		MOVF	PwmC,W
		BTFSC	STATUS,Z	; Si le servo est dsactiv,
		MOVLW	D'128'		; le TMR0 est arbitrairement mis  128
		MOVWF	TMR0

; Phase suivante
		MOVLW	1<<CPhase3
		MOVWF	Phase

		GOTO	LFinPhase

LPhase3
; Dure de 2,062 ms  0,054 ms
; Inversement proportionnel au pwm courant ( complment de la phase 2 )

		MOVF	PwmC,W
		BTFSC	STATUS,Z
		MOVLW	D'128'
		SUBLW	0xFF
		MOVWF	TMR0

; Phase suivante
		MOVLW	1<<CPhase1
		MOVWF	Phase


; Cette suite d'instruction aurait pu tre remplac par
; un simple CLRF PORTB, mais cette mthode aurait provoqu
; une diffrence entre les 8 servos.

		MOVF	IndPwm,W
		MOVWF	IndTmp

		BTFSC	STATUS,Z
		BCF	PORTB, CSBPortSM0

		DECF	IndTmp,F
		BTFSC	STATUS,Z
		BCF	PORTB, CSBPortSM1

		DECF	IndTmp,F
		BTFSC	STATUS,Z
		BCF	PORTB, CSBPortSM2

		DECF	IndTmp,F
		BTFSC	STATUS,Z
		BCF	PORTB, CSBPortSM3

		DECF	IndTmp,F
		BTFSC	STATUS,Z
		BCF	PORTB, CSBPortSM4

		DECF	IndTmp,F
		BTFSC	STATUS,Z
		BCF	PORTB, CSBPortSM5

		DECF	IndTmp,F
		BTFSC	STATUS,Z
		BCF	PORTB, CSBPortSM6

		DECF	IndTmp,F
		BTFSC	STATUS,Z
		BCF	PORTB, CSBPortSM7

; Servo suivant

		INCF	IndPwm,F
		BTFSC	IndPwm,3
		CLRF	IndPwm

		;GOTO	LFinPhase

LFinPhase

;***** Restitution contexte
		MOVF    status_temp,W	; restore STATUS
		MOVWF	STATUS		; /
		SWAPF   w_temp,F  	; restore W
		SWAPF   w_temp,W        ; / 

		RETFIE                  ; return aprs l'interruption

;**********************************************************************
;
;    Programme Main
;
;**********************************************************************

main

; Init de la RAM ( de 0C  4F ) Tout a 0
	MOVLW	0x0C
	MOVWF	FSR
LBoucleRam
	CLRF	INDF
	INCF	FSR,F
	MOVLW	0x50
	SUBWF	FSR,W
	BTFSS	STATUS,C
	GOTO	LBoucleRam

; Initialisation des variables ( Pour une valeur autre que 0 )

	MOVLW	1<<CPhase1
	MOVWF	Phase

; Initialisation des ports d'E/S, du Precale et Interruption timer

	CLRWDT			; Clear le WDT et le precale

	CLRF	PORTB		; Initialise le port B
	CLRF 	PORTA		; Initialise le port A

	BSF	STATUS, RP0	; Bank 1 slectionne

	MOVLW	B'00000000'	; Port RB7-0 en sortie
	MOVWF	TRISB		; /

	MOVLW	B'00011111'	; Port RA4-RA0 en entre
	MOVWF	TRISA		;	/

	MOVLW	CPrecale1
	MOVWF	OPTION_REG

	BCF 	STATUS,	RP0	; Bank 0 slectionne

	CLRF	TMR0		; Init du Timer
	MOVLW   B'10100000'     ; Active Interruption timer
	MOVWF   INTCON          ;       /

;======================================================================
; Fin de l'init du pic.
; Dbut de la boucle de lecture de la liaison srie.
;======================================================================

BoucleLectSerie

; Attente Entte = 255

	CALL	FGet

BLectSerie2
	MOVLW	D'255'
	SUBWF	Donnee,W
	BTFSS	STATUS,Z
	GOTO	BoucleLectSerie

; Lecture et Vrification adresse carte
; L'adresse est envoye aprs l'entte et doit correspondre
; a l'adresse de la carte prsent sur les ports RA2  RA4

	CALL	FGet

	MOVF	PORTA,W
	ANDLW	B'00011100'
	MOVWF	AdrCarte
	BCF	STATUS,C
	RRF	AdrCarte,F
	RRF	AdrCarte,F

	MOVF	AdrCarte,W
	SUBWF	Donnee,W
	BTFSS	STATUS,Z
	GOTO	BLectSerie2


; Lecture et Vrification du numro du servo
; Le numro du servo doit tre compris entre 0 et 7

	CALL	FGet

	MOVLW	D'8'
	SUBWF	Donnee,W
	BTFSC	STATUS,C
	GOTO	BLectSerie2

	MOVF	Donnee,W
	MOVWF	AdrServo

; Lecture position servo
; La position du servo est comprise entre 4 et 255,
; ou 0 pour le dsactiver 

	CALL	FGet

	MOVF	Donnee,W
	MOVWF	ValServo
	BTFSC	STATUS,Z
	GOTO 	BLectSerie3

	MOVLW	CPwmMin
	SUBWF	Donnee,W
	BTFSC	STATUS,C
	GOTO	BLectSerie3
	MOVLW	CPwmMin
	MOVWF	ValServo

BLectSerie3

; Affectation position servo

	MOVLW	PwmSM0
	ADDWF	AdrServo,W
	MOVWF	FSR
	MOVF	ValServo,W
	MOVWF	INDF

; fin de la boucle
	
	GOTO	BoucleLectSerie

;======================================================================

; fin du main

;**********************************************************************
;
;    Fonction FGet
; 	Lecture d'un octet sur la liaison srie.
; 	Au format : 1 start, 8 bits, 1 stop  2400 Bauds
;
;**********************************************************************
FGet
	MOVLW	D'8'
	MOVWF	NbrDonnee

; Attendre bit start :
; Dtection front descendant (attendre niveau 1, puis niveau 0 )

	BTFSS	PORTA,CEAPortSerie
	GOTO	$-1
	BTFSC	PORTA,CEAPortSerie
	GOTO	$-1

; Pause
	CALL	PauseGet

BFGet1
	CALL	PauseGet
	CALL	PauseGet

	BCF	STATUS,C
	BTFSC	PORTA,CEAPortSerie
	BSF	STATUS,C
	RRF	Donnee,F

	DECFSZ	NbrDonnee,F
	GOTO	BFGet1

	RETURN

; 1/2 pause entre 2 bits  2400  bauds
; La valeur thorique est de 208s ( 1/2 * 1000000 / 2400 )
; Mais la pause n'est que de 198s pour rester synchronis
; car 5% du cpu est utilis pour grer les signaux des servomoteurs
; lors des interruptions.

PauseGet
	MOVLW	D'64'	; 198s = 64 * 3s + 6s
	MOVWF	PauseV
	DECFSZ	PauseV,F
	GOTO	$-1
	RETURN

;**********************************************************************
;
;    Fin
;
;**********************************************************************

	END

